xquery version "3.1";
(:~
 : This modules handles the automatic enrichment of a whitelist file that is
 : used for different purposes in the BdN-project:
 :
 : * It serves a place to look up which files belong to the respective edition
 : * It provides data about the publication status of an edition and assures
 : correct availability on the live server [TODO]
 :
 : @author Michelle Weidling
 : @author Stefan Hynek
 : @version 2.1
 :)
module namespace whitelist="https://bdn-edition.de/ns/whitelist";

declare namespace bdn="http://www.bdn-edition.de/ns/1.0";
declare namespace tei="http://www.tei-c.org/ns/1.0";
declare namespace xi="http://www.w3.org/2001/XInclude";

import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace bdn-utils="http://bdn-edition.de/ns/utils" at "utils.xqm";
import module namespace functx="http://www.functx.com";

declare function whitelist:main($author as xs:string, $log as xs:string)
{
  bdn-utils:add-log-entry($log, "Start creating the whitelist."),
  whitelist:prepare-whitelist(),
  whitelist:add-files($author, $log),
  whitelist:add-split-files($author, $log)
};


(:~
 : assures that the whitelist is in the right place and has the correct filename.
 :
 : @author Michelle Weidling
 : @author Stefan Hynek
 : @version 2.1
 :)
declare function whitelist:prepare-whitelist() as item()*
{
    bdn-utils:assure-dir-available("aux"),
    if (doc-available($config:aux || "whitelist.xml")) then
        ()
    else
        xmldb:copy-resource($config:sade-data, "3qrcn.xml", $config:aux, "whitelist.xml")
};


(:~
 : Adds information about the TextGrid URIs that belong to an author.
 :
 : @author Michelle Weidling
 : @version 2.0
 : @param $author The current author, e.g. "noesselt"
 : @param $uris A sequence of textgrid uris that belong to an author, e.g. ("textgrid:12345", "textgrid:abcde")
 : @param $log The path to the current log file, e.g. $config:app-root || "logs/griesbach-log.xml"
 :)
declare function whitelist:add-files($author as xs:string, $log as xs:string)
as empty-sequence()
{
    let $author-entry := whitelist:get-author-entry($author)
    return
        if ($author-entry/following-sibling::bdn:files/bdn:file) then
            ()
        else
            let $uris := bdn-utils:get-structure-file($author)/tei:TEI//xi:include/@href => distinct-values()
            let $files :=
                for $uri in $uris return
                    <file xmlns="http://www.bdn-edition.de/ns/1.0">{$uri}</file>
            return
                (update insert $files into $author-entry/following-sibling::bdn:files,
                bdn-utils:add-log-entry($log, "Added TG-uris to whitelist"))
};

(:~
 : Adds the URIs of all files that have been created in xml-utils:semantic-splitting.
 :
 : @author Michelle Weidling
 : @version 2.0
 : @param $author The current author, e.g. "noesselt"
 : @param $log The path to the current log file, e.g. $config:app-root || "logs/griesbach-log.xml"
 :)
declare function whitelist:add-split-files($author as xs:string,
$log as xs:string) as empty-sequence()
{
    let $author-entry := whitelist:get-author-entry($author)
    return
        if ($author-entry/following-sibling::bdn:split-files/bdn:file) then
            ()
        else
            let $split-files-entry := $author-entry/following-sibling::bdn:split-files
            return
                let $uris := bdn-utils:retrieve-author-specific-uris($author)
                let $xml-fragments :=
                    for $uri in $uris return
                        <file xmlns="http://www.bdn-edition.de/ns/1.0">
                            {$uri}
                        </file>
                let $split-files-xml-fragment :=
                    <split-files xmlns="http://www.bdn-edition.de/ns/1.0">
                        {$xml-fragments}
                    </split-files>
                return
                    (
                      if ($split-files-entry) then
                          update replace $split-files-entry with $split-files-xml-fragment
                      else
                          update insert $split-files-xml-fragment following $author-entry/following-sibling::bdn:files,
                      bdn-utils:add-log-entry($log, "Added split files to whitelist")
                    )
};


(:~
 : Adds information about whether an index type that occurs in an edition.
 :
 : @author Michelle Weidling
 : @version 2.0
 : @param $indices A map with information about which type occurs in a work, e.g. map {"bible-index": true(), "persons-index": true(), "subjects-index": true(), "classics-index": false()}
 : @param $author The current author, e.g. "noesselt"
 :)
declare function whitelist:add-index-type($indices as map(*), $author as xs:string)
as xs:string*
{
    let $author-entry := whitelist:get-author-entry($author)
    return
        if ($author-entry/following-sibling::bdn:index) then
            ()
        else
            let $occuring :=
                function($type, $occurs)
                {
                    if ($occurs) then
                        <index xmlns="http://www.bdn-edition.de/ns/1.0" type="{$type}" />
                    else
                        ()
                }
            let $occuring-types :=
                function($type, $occurs)
                {
                    if ($occurs) then
                        $type
                    else
                        ()
                }
            let $occuring-indices := map:for-each($indices, $occuring)
            let $update := update insert $occuring-indices following $author-entry
            return map:for-each($indices, $occuring-types)
};

(:~
 : Adds the items that should appear in the TOC / Bandübersicht /work overview
 : to whitelist.xml.
 :
 : @author Michelle Weidling
 : @version 2.0
 : @param $author The current author, e.g. "noesselt"
 : @param $log The path to the current log file, e.g. $config:app-root || "logs/griesbach-log.xml"
 :)
declare function whitelist:add-nav-entries($author as xs:string,
$log as xs:string) as empty-sequence()
{
  let $author-entry := whitelist:get-author-entry($author)
  return
    (: nav-entries exist :)
      if ($author-entry/following-sibling::bdn:nav-entries) then
          ()
    (: nav-entries do not exist :)
      else
          let $body-structure := $author-entry/following-sibling::bdn:body-structure
          let $full-file := doc($config:full-xmls || $author || "-full.xml")
          let $id-file:= doc($config:aux || "id-file.xml")
          let $no-of-volumes := count($body-structure/bdn:volume)

          let $nav-fragments :=
              if ($no-of-volumes gt 1) then
                  for $volume in $body-structure/bdn:volume return
                      let $no := xs:integer($volume/@n)
                      let $text := $full-file//tei:text[$no]
                      let $head := $text//tei:supplied[@reason = "toc-title"][1]/string()

                      return
                          element bdn:entry
                          {
                              if ($volume/@scope) then
                                  let $id := replace($volume/@scope, "#", "")
                                  let $uri := $id-file//bdn:id[./string() = $id]/parent::bdn:file/@uri
                                  return
                                      attribute source {$uri}
                              else
                                  (),
                              $head,
                              if ($volume/child::*[@scope]) then
                                  whitelist:make-nav-units($volume/child::*, $full-file, $id-file, true())
                              else if ($volume/child::*[@n]) then
                                  whitelist:make-nav-units($volume/child::*, $full-file, $id-file, false())
                              else
                                  ()
                          }

              else
                  if ($author = "less") then
                      whitelist:make-nav-entries-for-less($full-file, $id-file, $body-structure/bdn:volume)
                  else
                      let $text := $full-file//tei:text[1]
                      let $head := $text//tei:supplied[@reason = "toc-title"][1]/string()

                      return
                          if ($body-structure/bdn:volume/child::*) then
                              whitelist:make-nav-units($body-structure/bdn:volume/child::*, $full-file, $id-file, false())
                          else
                              element bdn:entry
                              {
                                  if ($body-structure/bdn:volume/@scope) then
                                      let $id := replace($body-structure/bdn:volume/@scope, "#", "")
                                      let $uri := $id-file//bdn:id[./string() = $id]/parent::bdn:file/@uri
                                      return
                                          attribute source {$uri}
                                  else
                                      (),
                                  $head
                              }

  let $nav-entry :=
          <nav-entries xmlns="http://www.bdn-edition.de/ns/1.0">
              {$nav-fragments}
          </nav-entries>
  return
      (
        update insert $nav-entry following $author-entry/following-sibling::bdn:split-files,
        bdn-utils:add-log-entry($log, "Added headings for nav entries to whitelist")
      )
};


(:~
 : Takes the children of a bdn:volume and creates (sub-)entry elements for the
 : whitelist.
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $units The children of a bdn:volume element
 : @param $full-file The full TEI/XML file of a given author, e.g. doc("/db/apps/sade/resources/full-xmls/griesbach.xml")/*
 : @param $id-file The opened ID file
 : @param $isSubentry true() creates a bdn:subentry, false() a bdn:entry element
 :)
declare function whitelist:make-nav-units($units as node()+, $full-file as node(),
$id-file as node(), $isSubentry as xs:boolean) as element(*)*
{
    for $unit in $units return
        (: in this case the current unit has no child nodes :)
        if ($unit/@scope
        and not($unit/child::*[@scope])) then
            let $ids := tokenize($unit/@scope)
            let $ids :=
                for $id in $ids return
                    replace($id, "#", "")

            let $heads :=
                for $id in $ids return
                    $full-file//*[@xml:id = $id]//tei:supplied[@reason = "toc-title"][1]/string()
            let $heads-no := count($heads)

            let $uris :=
                for $id in $ids return
                    $id-file//bdn:id[./string() = $id]/parent::bdn:file/@uri

            return
                for $iii in 1 to $heads-no return
                    if ($isSubentry) then
                        <subentry xmlns="http://www.bdn-edition.de/ns/1.0" source="{$uris[$iii]}">
                            {
                                $heads[$iii]
                            }
                        </subentry>
                    else
                        <entry xmlns="http://www.bdn-edition.de/ns/1.0" source="{$uris[$iii]}">
                            {
                                $heads[$iii]
                            }
                        </entry>
        else if ($unit[self::bdn:front]) then
            let $author := $unit/ancestor::bdn:work/bdn:author/string()
            let $vol-no := $unit/parent::*/@n
            let $uri := $author || "-" || $vol-no || "-" || "front"
            return
                if ($isSubentry) then
                    <subentry xmlns="http://www.bdn-edition.de/ns/1.0" source="{$uri}">
                        Vorrede(n)
                    </subentry>
                else
                    <entry xmlns="http://www.bdn-edition.de/ns/1.0" source="{$uri}">
                        Vorrede(n)
                    </entry>

        else if ($unit[self::bdn:back]) then
            let $author := $unit/ancestor::bdn:work/bdn:author/string()
            let $vol-no := $unit/parent::*/@n
            let $uri := $author || "-" || $vol-no || "-" || "back"
            return
                if ($isSubentry) then
                    <subentry xmlns="http://www.bdn-edition.de/ns/1.0" source="{$uri}">
                       Druckfehler
                    </subentry>
                else
                    <entry xmlns="http://www.bdn-edition.de/ns/1.0" source="{$uri}">
                       Druckfehler
                    </entry>

        (: in this case the current unit has child nodes :)
        else
            let $id := replace($unit/@scope, "#", "")
            let $head := $full-file//*[@xml:id = $id]//tei:supplied[@reason = "toc-title"][1]/string()
            return
                <subentry xmlns="http://www.bdn-edition.de/ns/1.0">
                {
                    $head,
                    for $child in $unit/child::* return
                        whitelist:make-nav-units($child, $full-file, $id-file, true())
                }
                </subentry>
};

(:~
 : Leß has a very loose structure that need manual splitting for the navigation;
 : else we'd have around 50 sermons as separate entries in the work overview.
 :
 : To make this a bit more handy, we manually divided the sermons into groups
 : (given in whitelist.xml) treat these groups as substructure (which actually
 : isn't there in the source).
 :
 : @author Michelle Weidling
 : @version 1.0
 : @param $full-fille The full TEI/XML file of a given author, e.g. doc("/db/apps/sade/resources/full-xmls/griesbach.xml")/*
 : @param $id-file The opened ID file
 : @param $volume Leß's volume as bdn:volume element
 : @return A bdn:entry element for each bdn:sermon-range
 :)
declare function whitelist:make-nav-entries-for-less($full-file as node(),
$id-file as node(), $volume as element(bdn:volume)) as element(bdn:entry)+
{
    let $ranges := $volume/child::*
    return
        for $range in $ranges return
        if ($range[self::bdn:front]) then
            let $uri := "less-1-front"
            return
                <entry xmlns="http://www.bdn-edition.de/ns/1.0" source="{$uri}">
                    Vorrede(n)
                </entry>

        else if ($range[self::bdn:back]) then
            let $uri := "less-1-back"
            return
                <entry xmlns="http://www.bdn-edition.de/ns/1.0" source="{$uri}">
                    Anhang
                </entry>
        else
            <entry xmlns="http://www.bdn-edition.de/ns/1.0">
                {
                    $range/@title/string(),
                    whitelist:make-nav-units($range/child::*, $full-file, $id-file, true())
                }
            </entry>
};

(:~
 : Adds a hint whether structural variance appears on chapter level. In this
 : case we have to deal with these chapters differently in the navbar.
 :
 : @author Michelle Weidling
 : @version 2.0
 : @since 2.0
 : @param $author The current author, e.g. "noesselt"
 : @param $log The path to the current log file, e.g. $config:app-root || "logs/griesbach-log.xml"
 :)
declare function whitelist:mark-sv-on-chapter-level($author as xs:string,
$log as xs:string) as empty-sequence()
{
    let $author-entry := whitelist:get-author-entry($author)
    let $full-file := doc($config:full-xmls || $author || "-full.xml")
    let $part-chapter-with-sv := $full-file//tei:div[@type = ("part", "chapter") and parent::tei:lem/following-sibling::tei:rdg[@type = "missing-structure"]]

    return
        if ($part-chapter-with-sv) then
            let $structural-variances :=
                for $node in $part-chapter-with-sv return
                    let $id := $node/@xml:id/string()
                    let $printing := $node/../following-sibling::tei:rdg[@type = "missing-structure"]/@wit => replace("#", "")
                    let $file := doc($config:aux || "id-file.xml")//bdn:id[./string() = $id]/parent::*/@uri/string()
                    let $corresp-nav-entry := $author-entry/..//bdn:entry[@source = $file]
                    let $updated := functx:add-attributes($corresp-nav-entry, xs:QName("omitting-printing"), $printing)

                    return
                        update replace $corresp-nav-entry with $updated

            return
                bdn-utils:add-log-entry($log, "Added occurences of structural variance on chapter level to whitelist")
        else
            ()
};


declare function whitelist:create-sv-navbar($sv-elements as node()+) as empty-sequence()
{
    ()
};

declare function whitelist:get-whitelist() as node()*
{
  doc($config:aux || "whitelist.xml")
};


declare function whitelist:get-author-entry($author as xs:string)
as element(bdn:author)
{
  let $whitelist := whitelist:get-whitelist()
  return $whitelist//bdn:author[./string() = $author]
};
