xquery version "3.1";

(:~
 : This module creates an SVG image of call graph of
 :
 : * a given module
 : * the modules of a given collection
 : * the modules of the whole SADE app
 :
 : A module is defined as a file that ends with '.xqm'.
 : While the SVG image is stored to /db/apps/sade/modules/callgraph/svg,
 : the intermediate GraphML representation of the call graph is saved to
 : /db/apps/sade/modules/callgraph/graphml and could be re-used in other
 : applications.
 :
 : This module was roughly inspired by the graphviz app for eXist-db which can be
 : found at https://github.com/KitWallace/graphviz.
 :
 : @author Michelle Weidling
 : @since v3.1.1
 : @see https://github.com/KitWallace/graphviz
 :
 : :)


module namespace callgraph="https://sade.textgrid.de/ns/callgraph";

import module namespace config="https://sade.textgrid.de/ns/config" at "../config.xqm";
import module namespace inspect="http://exist-db.org/xquery/inspection";
import module namespace inspect2graphml="https://sade.textgrid.de/ns/inspect2graphml" at "inspect2graphml.xqm";

declare namespace graphml="http://graphml.graphdrawing.org/xmlns";
declare namespace xsi="http://www.w3.org/2001/XMLSchema-instance";

declare variable $callgraph:xslt := $config:app-root || "modules/callgraph/graphml2dot.xsl";
declare variable $callgraph:colors :=
    ("#90b0d4", "#e6f2ff", "#eee6ff", "#ffffe6",
    "#f9ecec", "#bfe0ba", "#bfd2e3", "#dbaddb",
    "#decbb2", "#e0b7b6", "#d9f09e", "#f5a1a4",
    "#faf0a5", "#8acca0", "#cfb88c", "#d18ecb");

(:~
 : Creates an SVG file of one or more modules depending on the given $option.
 : Possible options are:
 : * an URI of a single module, e.g. "/db/apps/sade/modules/callgraph/tests/xqm/inspect2graphml-testmodule.xqm"
 : * a collection URI, e.g. "/db/apps/sade/modules/callgraph/tests/xqm/"
 : * "full" which will consider all *.xqm in the SADE app
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @param $option Selects the modules which should be displayed in the output SVG file
 : @param $filename The filename without extension, e.g. "output"
 : @return Message that indicates a successfull transformation
 : @error The given option is invalid
 : @error The given filename is invalid
 :  :)
declare function callgraph:main($option as xs:string, $filename as xs:string) as xs:string {
    if(ends-with($option, ".xqm")
    or $option = "full"
    or xmldb:collection-available($option)) then
        let $filename :=
            if(contains($filename, ".")) then
                error( QName("https://sade.textgrid.de/ns/app", "CALLGRAPH01"),
                "Invalid filename '" || $filename || "'. Please enter a filename without file extension.")
            else
                $filename
        let $dot := callgraph:get-dot($option, $filename)
        let $svg-output := callgraph:get-svg($dot)
        let $svg := callgraph:clear-svg($svg-output, $filename)

        return "Call graph creation successfull."
    else
        error( QName("https://sade.textgrid.de/ns/app", "CALLGRAPH02"),
            "Invalid option '" || $option || "'. Please enter a base URI, a collection path or 'full' as argument.")
};


(:~
 : Gets a DOT representation of the given module, the modules of a collection or
 : all modules of the SADE app.
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @see https://en.wikipedia.org/wiki/DOT_(graph_description_language)
 : @param $option Selects the modules which should be displayed in the output SVG file (cf. callgraph:main)
 : @return The DOT representation of the call graph
 : @error GraphML couldn't be transformed to DOT
 : @error The DOT file is empty
 :  :)
declare function callgraph:get-dot($option as xs:string, $filename as xs:string)
as xs:string {
    let $graphml := inspect2graphml:main($option, $filename)
    let $dot :=
        try {
            (: $graphml[1] is graphml:graphml,
            $graphml[2] is path to stored XML :)
            callgraph:transform2dot($graphml[1])
            => string-join("")
        } catch * {
            error( QName("https://sade.textgrid.de/ns/app", "CALLGRAPH03"),
            "Could not transform GraphML to DOT.")
        }

    return
        if(normalize-space($dot) != "") then
            $dot
        else
            error( QName("https://sade.textgrid.de/ns/app", "CALLGRAPH04"),
            "DOT file is empty.")
};


(:~
 : Transforms the given GraphML elements to DOT.
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @see https://en.wikipedia.org/wiki/DOT_(graph_description_language)
 : @param $nodes The GraphML nodes generated by inspect2graphml:main
 : @return The DOT representation of the call graph
 :  :)
declare function callgraph:transform2dot($nodes as element()*) as node()* {
    for $node in $nodes return
        typeswitch ($node)

        case element(graphml:graphml) return
            (text{"digraph g {"},
            callgraph:transform2dot($node/node()),
            text{"}"})

        case element(graphml:graph) return
            if($node/@id = "base-graph") then
                callgraph:transform2dot($node/node())
            else
                (text{"node [shape=box, style=filled, color="""},
                text{$callgraph:colors[count($node/preceding::graphml:graph) + 1]},
                text{"""];"},
                callgraph:transform2dot($node/node()))

        case element(graphml:node) return
            if($node/parent::graphml:graph[@id = "base-graph"]) then
                callgraph:transform2dot($node/node())
            else
                (text{""""},
                text{$node/@id},
                text{""" [shape=box];"})

        case element(graphml:edge) return
            (text{""""},
            text{$node/@source},
            text{""""},
            text{" -> "},
            text{""""},
            text{$node/@target},
            text{""";"})

        default return
            callgraph:transform2dot($node/node())
};



(:~
 : Creates an SVG file based of a given DOT file.
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @see https://en.wikipedia.org/wiki/DOT_(graph_description_language)
 : @param $dot A graph denoted in DOT
 : @return The lines of the SVG file as element(line)
 : @error DOT couldn't be transformed to SVG
 : @error The SVG file is empty
 :  :)
declare function callgraph:get-svg($dot as xs:string) as element(line)+ {
    let $cmd := $config:file//*[@key = "dot-cmd"]/string()
    let $target-dir := $config:file//*[@key = "target-dir"]/string()
    let $options :=
        <options>
            <workingDir>{$target-dir}</workingDir>
            <stdin><line>{$dot}</line></stdin>
        </options>
    let $output :=
        try {
            process:execute(($cmd, "-Tsvg"), $options)
        } catch * {
            error( QName("https://sade.textgrid.de/ns/app", "CALLGRAPH05"),
            "Could not transform the given DOT file to SVG.")
        }
    return
        if($output//stdout/line) then
            $output
        else
           error( QName("https://sade.textgrid.de/ns/app", "CALLGRAPH06"),
            "SVG output empty.")
};


(:~
 : Since the output of callgraph:get-svg consists of multiple line elements that
 : contain the XML representation of the SVG, this function extracts the SVG data
 : and creates a valid SVG file.
 :
 : @author Michelle Weidling
 : @since v3.1.0
 : @param $svg Multiple line elements that hold the SVG data
 : @param $filename The name of the generated SVG file
 : @return The path to the stored SVG file
 :  :)
declare function callgraph:clear-svg($svg as element(execution), $filename as
xs:string) as xs:string {
    let $lines := $svg//line
    let $content :=
        for $line in subsequence($lines, 7) return
            $line/string()
    let $cleared-svg := string-join($content, "")
    return
        xmldb:store($config:app-root || "modules/callgraph/svg/", $filename || ".svg", $cleared-svg)
};
